package lib; //Failed.java

//History:
//2000-07-13  Knabe  Efficiently reimplemented: printCompactStackTrace
//2000-07-09  Knabe  Retrofitted to JDK 1.1: Uses Vector instead of List
//2000-06-29  Knabe  Class Exc split into Exc and Failed


/** Fully diagnostic Java exception with parameters and cause chain.
  The exception parameters are inherited from Exc.
  This class serves for the framework user as a base class for defining
  Failed-exceptions, which have the following benefits: <UL>
    <LI> coherent usage always for unexpected failure of an operation
    <LI> reduce the amount of exceptions to specify in the throws-clause
         from many unexpected exceptions to only one
    <LI> provide the failure of an operation with a message text,
         message parameters and a chain of the unexpected causing exceptions.
  </UL>
  <P>
  Naming convention:
    All user-defined exception classes derived from Failed should have a name
    ending in Failed. The pattern is: <PRE>
      class OperationnameFailed extends lib.Failed { ... }
    </PRE>
  @author Christoph Knabe, TFH Berlin, 1999-2000 Copyrighted
*/

public class Failed extends Exc {


/**The string used to delimit the stack trace part of a causing Exception from
  the following stack trace of the caused Failed-exception.
*/
private static final String _wasCausing = "WAS CAUSING:";

/**Constructs a Failed-exception, giving the causing Exception object for
  providing the diagnostics causer chain and the message parameters as an Object[]

  <PRE><CODE>

  try{
    many calls to service methods ...
  } catch(Exception ex){ //all these are unexpected
      throw new OperationFailed(ex, additional diagnostic parameters );
    }
  //end try

  </PRE></CODE>
*/
public Failed(final Exception i_cause, final Object[] i_parameters){
  super(i_parameters);
  _cause = i_cause;
}

/** Constructs a Failed-exception, giving only the causing Exception object.
*/
public Failed(final Exception i_cause){
  _cause = i_cause;
}

/**Result: The Exception, which caused the program to throw this exception.
  I.e. the Exception, which was redefined into this Exc object.
*/
public final Exception cause(){return _cause;}

/** Result: the causing Exception of i_throwable, if there is one,
* otherwise null.
*/
public final static Exception cause(final Throwable i_throwable){
  return i_throwable instanceof Failed
    ? ((Failed)i_throwable).cause()
    : null
  ;
}

/**Prints the chained, redundancy-free stack traces of this <code>Failed</code>
  object to the standard error stream System.err
  @see #printStackTrace(java.io.PrintWriter)
  @see java.lang.System#err
*/
public void printStackTrace(){
  printStackTrace(System.err);
}

/**Prints the chained, redundancy-free stack traces of this <code>Failed</code>
  object to io_printer.
* @see #printStackTrace(java.io.PrintWriter)
*/
public void printStackTrace(final java.io.PrintStream io_printer) {
  printStackTrace(new java.io.PrintWriter(io_printer, true));
}

/** Prints the chained, redundant stack traces of this <code>Failed</code>
  object to io_printer.
  The output contains the name, message parameters and stack trace first of
  the causing <code>Exception</code> object, then of this <code>Exc</code>
  object to the specified print writer. Users should usually not call this
  method directly, but printCompactStackTrace instead in order to avoid
  redundant and useless lines in the stack traces.
  @see #printCompactStackTrace
*/
public void printStackTrace(final java.io.PrintWriter io_printer) {
  //simple efficient code: leaves all redundant lines:
  if(_cause!=null){
    _cause.printStackTrace(io_printer);
    io_printer.println(_wasCausing);
  }
  super.printStackTrace(io_printer);
  io_printer.flush();
}


/** Prints the stack trace of i_throwable and all its chained cause exceptions
  to io_printer. All redundant location lines therein are suppressed.
  As well all location lines, which are also contained in the stack trace of
  i_here, are suppressed.
  This is useful, e.g. for reporting an error in a Swing-ActionListener,
  because the stack trace of an actionPerformed-method contains 19 lines
  inside of the AWT/Swing-Package. These lines are not useful for error
  searching, as they are always the same in any Swing application.
  Usually you should not directly call this method, but a report-method in
  class Msg.
  @param io_printer Where to print the compactified stack trace
  @param i_reportee The exception to report, possibly containing a chain of
    nested cause exceptions
  @param i_here An dummy exception, which should have been newly created there,
    where the reporting for i_reportee is invoked
  @see Msg#report(java.awt.Frame,Throwable)
*/
public static void printCompactStackTrace(
  final StringBuffer io_buffer, final Throwable i_reportee,
  final Exception i_here
){
  //prepare out writer:
  final java.io.StringWriter sw = new java.io.StringWriter();
  final java.io.PrintWriter out = new java.io.PrintWriter(sw, true);
  //final java.io.PrintWriter err = new java.io.PrintWriter(System.err, true);

  //collect stack traces onto out writer:
  i_reportee.printStackTrace(out);
  out.println(_wasCausing);
  i_here.printStackTrace(out);
  final String tracesString = sw.toString();
  //System.err.println("tracesString:");System.err.println(tracesString);//debug
  final java.util./*String*/Vector tracesLines
  = _splitIntoLines(tracesString);

  //nullify redundant trace trail lines:
  final int[] traceEndIndices = _getTraceEndIndices(tracesLines);
  //_println("traceEndIndices", traceEndIndices); //debug
  for(int i=1; i<traceEndIndices.length; i++){
    _nullifyCommonTraceTail(
      tracesLines, traceEndIndices[i-1], traceEndIndices[i]
    );
  }
  _shortenLastTrace(
    tracesLines, traceEndIndices[traceEndIndices.length-2]+1
  );
  _printLines(io_buffer, tracesLines);
}//printCompactStackTrace

/**Splits i_string into internized line strings by BufferedReader.readln().
  @return The lines in the sequence of storage in i_string as a Vector of
          String-s.
*/
private static java.util./*String*/Vector _splitIntoLines(final String i_string){
  final java.io.BufferedReader reader = new java.io.BufferedReader(
    new java.io.StringReader(i_string)
  );
  final java.util./*String*/Vector result = new java.util.Vector();
  for(;;){
    String line; //Warum geht es hier nicht final?
    try{line = reader.readLine();}
      catch(java.io.IOException ex){line=null;}
    //end try
    if(line==null){break;}
    result.addElement(line.intern());
  }
  return result;
}

/**Result: The indices of the elements of i_tracesLines, which are immediately
  before the occurences of the String _wasCausing in i_tracesLines.
  Require: In i_traces are several stack traces, split into lines; the traces
    are separated by the String _wasCausing.
*/
private static int[] _getTraceEndIndices(final java.util.Vector i_tracesLines){
  final int numberOfTraces
  = _numberOfOccurences(i_tracesLines, _wasCausing) + 1;
  final int[] result = new int[numberOfTraces];
  final int size = i_tracesLines.size();
  int resultIndex = 0;
  for(int i=0; i<size; i++){
    if(i_tracesLines.elementAt(i) == _wasCausing){
      result[resultIndex++] = i-1; //line before _wasCausing
    }
  }
  result[resultIndex++] = size-1; //last line
  return result;
}

/**Result: The number, how often i_object occurs in i_vector,
  measured by the identity comparison
*/
private static int _numberOfOccurences
(final java.util.Vector i_vector, Object i_object){
  final int size = i_vector.size();
  int result = 0;
  for(int i=0; i<size; i++){
    if(i_vector.elementAt(i) == i_object){result++;}
  }
  return result;
}

/**Nullifies redundant entries in io_vector.
  Nullifies all entries in io_vector
  starting from i_nullifyEndIndex inclusively downwards,
  which are identically contained in the same
  sequence as entries starting from i_compareEndIndex inclusively downwards.
*/
private static void _nullifyCommonTraceTail(
  final java.util.Vector io_vector,
  final int i_nullifyEndIndex, final int i_compareEndIndex
){
  for(int ni=i_nullifyEndIndex, ci=i_compareEndIndex; ;ni--,ci--){
    if(io_vector.elementAt(ni)!=io_vector.elementAt(ci)){break;}
    io_vector.setElementAt(null, ni);
  }
}//_nullifyCommonTraceTail

/**Shortens the Vector io_tracesLines by setting the element at index
  i_fromIndex to the abbreviation symbol "at ..." and nullifying all
  elements after this.
*/
private static void _shortenLastTrace(
  final java.util.Vector io_tracesLines, final int i_fromIndex
){
  io_tracesLines.setElementAt("\tat ...", i_fromIndex);
  final int size = io_tracesLines.size();
  for(int i=i_fromIndex+1; i<size; i++){
    io_tracesLines.setElementAt(null, i);
  }
}

/**Appends the non-null elements of i_lines each as a line to io_buffer*/
private static void _printLines(
  final StringBuffer    io_buffer,
  final java.util.Vector i_lines
){
  final int size = i_lines.size();
  for(int i=0; i<size; i++){
    final Object line = i_lines.elementAt(i);
    if(line!=null){io_buffer.append(line);io_buffer.append(MsgText.lineSeparator);}
  }
}

/*debug
private static void _println(final String i_name, final int[] i_array){
  System.err.print(i_name + ": ");
  for(int i=0; i<i_array.length; i++){
    System.err.print(i_array[i] + "  ");
  }
  System.err.println();
}
*/

/**The Exception object, that caused this Exc
  @see #cause()
*/
private Exception _cause;


}//Failed