package lib; //MsgText.java



//nderungshistorie:
//2000-07-24  Knabe  Umsetzung von Namensparametern mittels Reflection
//2000-07-07  Knabe  extrahiert aus Klasse Msg


/** Services for getting the message text for exceptions <UL>
  <LI> with parameters (subclasses of {@link Exc})
  <LI> with cause chain and parameters (subclasses of {@link Failed})
  <LI> and standard Java exceptions (other subclasses of java.lang.Throwable)
  </UL>
  @author Christoph Knabe, Berlin, Copyright 1999-2000
*/
public class MsgText {


private static final String _className = "MsgText";
private static final String _because = "CAUSE: ";

/**Returns the message for i_throwable with its causal chain.
  @return A usually multi-line String, in which the first line is the
    message for i_throwable and the further lines are the messages for the
    causing exceptions of i_throwable. The lines are separated by
    {@link #lineSeparator}. See {@link #getMessageLine} for a despription,
    how each exception is transformed into a message line.
*/
public static String getMessageChain(final Throwable i_throwable){
  final StringBuffer result = new StringBuffer(160);
  //Throwable-Meldungen eintragen:
  for(Throwable t=i_throwable; t!=null; t=Failed.cause(t)){
    if(t!=i_throwable){ //causing exception
      result.append(lineSeparator);
      result.append(_because);
    }
    result.append(getMessageLine(t));
  }
  return result.toString();
}


/**Result: A usually one-line message text for the exception
  i_throwable with inserted parameters.

  The message text for the exception is taken from the properties file
  "MsgText.properties" or one of the localized variants
  of it, e.g. "Msg_de_DE.properties", see the Java Tutorial
  <A HREF="http://java.sun.com/docs/books/tutorial/i18n/format/messageFormat.html">
  chapter on internationalization</A> at february 1999.
  <P>
  This file must contain definition lines of the form: <PRE>
     packageName.ExceptionName = message text pattern
  </PRE>or<PRE>
     packageName.ClassName$InnerExceptionName = message text pattern
  </PRE>with parameter substitutors in the message text pattern, e.g. <PRE>
     Copy$CopyFailed = The file "{0}" could not be copied to "{1}"
  </PRE>If i_throwable is a standard Java exception with one String as
  detailMessage, this is taken as message parameter {0}.
*/
public static String getMessageLine(final Throwable i_throwable){
  final String className = i_throwable.getClass().getName();
  _mut_tryLoadMessages();
  if(_messages==null){
    return i_throwable.toString();
  }
  try {
    String messageTextPattern = _messages.getString(className);
    final Object[] indexParams;
    final String detailMessage;
    final java.lang.reflect.Field[] namedParams
    = Exc._getPublicNostaticFields(i_throwable);
    if(namedParams!=null){ //with exception named parameters
      //transform into index parameters:
      messageTextPattern
      = _replaceNameByIndexPlaceholders(messageTextPattern,namedParams);
      indexParams = Exc._getFieldValues(i_throwable, namedParams);
    }else if(i_throwable instanceof Exc){ //may be index-parameter exception
      final Exc exc = (Exc)i_throwable;
      indexParams = exc.parameters();
    }else if((detailMessage=i_throwable.getMessage())!=null){
      //standard Java detailMessage as exception parameter {0}:
      indexParams = new Object[]{detailMessage};
    }else{
      indexParams=null;
    }
    if(indexParams==null){ //no exception parameters
      return messageTextPattern;
    }else{ //with exception parameters: format
      return java.text.MessageFormat.format(
        messageTextPattern, indexParams
      );
    }
  } catch(final Exception ex){
      return i_throwable.toString();
    }
  //end try
}//getMessageLine

/**Returns i_messagePattern, in which the name-placeholders
  are replaced by index-placeholders. Only placeholders for the names of the
  fields in i_param are replaced by the placeholder for the corresponding index.
*/
private static String _replaceNameByIndexPlaceholders(
  final String i_messagePattern,
  final java.lang.reflect.Field[] i_params
){
  final StringBuffer result = new StringBuffer(i_messagePattern.length());
  for(int ci=0; ci<i_messagePattern.length(); ci++){ //char loop
    final char c = i_messagePattern.charAt(ci);
    result.append(c);
    if(c=='{'){ //may be start of name parameter placeholder
      final int pi = _indexForMatchedParam(i_messagePattern, i_params, ci+1);
      if(0<=pi){ //???What if pi>9 ?
        //???Needs still translate the supress syntax. e.g.: {9-}
        result.append(pi);
        //skip param name:
        ci += i_params[pi].getName().length();
      }//if pi
    }//if c
  }//for ci
  return result.toString();
}//_replaceNameByIndexPlaceholders


private static final int _noMatch = -1;

/**Returns the index for the matched param, or _noMatch*/
private static int _indexForMatchedParam(
  final String i_messagePattern,
  final java.lang.reflect.Field[] i_params,
  final int i_startIndex
){
  final int patLen = i_messagePattern.length();
  if(i_startIndex >= patLen){return _noMatch;}

  for(int pi=0; pi<i_params.length; pi++){ //parameter loop
    final String name = i_params[pi].getName();
    final int namLen = name.length();
    final int endDelimI = i_startIndex + namLen;
    final char endDelim
    = endDelimI<patLen ? i_messagePattern.charAt(endDelimI) : '\0';
    if(i_messagePattern.regionMatches(i_startIndex, name, 0, namLen)
      //parameter name is matched
      && "},".indexOf(endDelim)>=0 //end delimiter is matched
    ){
      return pi;
    }
  }
  return _noMatch;
}

/**Returns the line separator for the actual platform.
  This is the line separator for the platform the Java virtual machine is
  running on. It is the value of the system property 'line.separator'.
*/
public static final String lineSeparator
= System.getProperty("line.separator");


//Message text access:

private static java.util.ResourceBundle _messages;
private static boolean                  _loadMessagesTried = false;

/**Tries once to load the ResourceBundle _className as message text file into
  _messages. If the loading failes, this method will leave _messages null, and
  will do nothing on further invocations.
*/
private static void _mut_tryLoadMessages(){
  if(_loadMessagesTried){return;}
  try{
    _loadMessagesTried = true;
    _messages = java.util.ResourceBundle.getBundle(_className);
  } catch(final Exception ex){ //not loadable
      ex.printStackTrace();
      _messages = null;
    }
  //end try
}//_mut_tryLoadMessages


}//MsgText
